import asyncio
import time

import pytest

from pyxxl.enum import executorBlockStrategy
from pyxxl.error import JobDuplicateError, JobNotFoundError, JobParamsError
from pyxxl.executor import Executor, JobHandler
from pyxxl.schema import RunData

job_handler = JobHandler()


@job_handler.register
async def pytest_executor_async():
    await asyncio.sleep(2)
    return "成功30"


@job_handler.register
def pytest_executor_sync():
    time.sleep(2)
    return "成功30"


@job_handler.register
async def pytest_executor_error():
    assert 1 == 2


HANDLER_NAMES = [
    "pytest_executor_async",
    "pytest_executor_sync",
]


@pytest.mark.asyncio
async def test_runner_not_found(executor: Executor, job_id: int):
    executor.reset_handler(job_handler)
    with pytest.raises(JobNotFoundError):
        await executor.run_job(
            RunData(
                **dict(
                    logId=211,
                    jobId=job_id,
                    executorHandler="not_found",
                    executorBlockStrategy=executorBlockStrategy.DISCARD_LATER.value,
                )
            )
        )
    await executor.graceful_close()


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_callback(executor: Executor, handler_name: str):
    executor.reset_handler(job_handler)
    executor.xxl_client.clear_result()
    data = RunData(
        **dict(
            logId=1,
            jobId=11,
            executorHandler=handler_name,
            executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
        )
    )
    await executor.run_job(data)
    data = RunData(
        **dict(
            logId=2,
            jobId=12,
            executorHandler="pytest_executor_error",
            executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
        )
    )
    await executor.run_job(data)
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(1) == 200
    assert executor.xxl_client.callback_result.get(2) == 500
    assert executor.xxl_client.callback_result.get(3) is None


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_cancel(executor: Executor, handler_name: str, job_id: int):
    executor.reset_handler(job_handler)
    cancel_job_id, ok_job_id = 1100, 1200
    base_data = dict(
        executorHandler=handler_name,
        executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
    )
    await executor.run_job(RunData(logId=cancel_job_id, jobId=cancel_job_id, **base_data))
    await executor.run_job(RunData(logId=ok_job_id, jobId=ok_job_id, **base_data))

    await executor.cancel_job(cancel_job_id, include_queue=False)
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(cancel_job_id) == 500
    assert executor.xxl_client.callback_result.get(ok_job_id) == 200
    # include_queue =True
    base_data.update({"jobId": job_id})
    executor.xxl_client.clear_result()
    cancel_log_id = job_id
    queue_log_id = job_id + 1
    await executor.run_job(RunData(logId=cancel_log_id, **base_data))
    await executor.run_job(RunData(logId=queue_log_id, **base_data))
    await executor.cancel_job(job_id, include_queue=True)
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(cancel_log_id) == 500
    assert executor.xxl_client.callback_result.get(queue_log_id) is None


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_SERIAL_EXECUTION(executor: Executor, job_id: int, handler_name: str):
    executor.xxl_client.clear_result()
    executor.reset_handler(job_handler)
    run_data = dict(
        jobId=job_id,
        executorHandler=handler_name,
        executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
    )
    await executor.run_job(RunData(logId=11, **run_data))
    await executor.run_job(RunData(logId=12, **run_data))
    await executor.run_job(RunData(logId=13, **run_data))
    assert executor.queue.get(job_id).qsize() == 2
    await executor.graceful_close()
    assert executor.queue.get(job_id).qsize() == 0
    assert executor.xxl_client.callback_result.get(13) == 200

    # max_queue_length
    for i in range(executor.config.task_queue_length + 1):
        await executor.run_job(RunData(logId=100 + i, **run_data))

    with pytest.raises(JobDuplicateError, match="discard"):
        await executor.run_job(RunData(logId=101 + i, **run_data))

    await executor.shutdown()


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_DISCARD_LATER(executor: Executor, job_id: int, handler_name: str):
    executor.reset_handler(job_handler)
    executor.xxl_client.clear_result()
    run_data = dict(
        jobId=job_id,
        executorHandler=handler_name,
        executorBlockStrategy=executorBlockStrategy.DISCARD_LATER.value,
    )
    await executor.run_job(RunData(logId=11, **run_data))
    with pytest.raises(JobDuplicateError):
        await executor.run_job(RunData(logId=21, **run_data))
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(11) == 200
    assert executor.xxl_client.callback_result.get(21) is None


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_COVER_EARLY(executor: Executor, job_id: int, handler_name: str):
    executor.reset_handler(job_handler)
    executor.xxl_client.clear_result()
    run_data = dict(
        jobId=job_id,
        executorHandler=handler_name,
        executorBlockStrategy=executorBlockStrategy.COVER_EARLY.value,
    )
    await executor.run_job(RunData(logId=40, **run_data))
    await executor.run_job(RunData(logId=41, **run_data))
    await executor.graceful_close()
    assert executor.xxl_client.callback_result.get(41) == 200
    assert executor.xxl_client.callback_result.get(40) == 500


@pytest.mark.asyncio
@pytest.mark.parametrize("handler_name", HANDLER_NAMES)
async def test_runner_OTHER(executor: Executor, job_id: int, handler_name: str):
    executor.reset_handler(job_handler)
    with pytest.raises(JobParamsError, match="unknown executorBlockStrategy"):
        for i in range(2):
            await executor.run_job(
                RunData(
                    logId=40 + i,
                    jobId=job_id,
                    executorHandler=handler_name,
                    executorBlockStrategy="OTHER",
                )
            )
    executor.xxl_client.clear_result()



import inspect
import logging
import os
from dataclasses import asdict, dataclass, field
from typing import Literal, Optional

from yarl import URL

from pyxxl.utils import get_network_ip, setup_logging

logger = logging.getLogger(__name__)


@dataclass
class ExecutorConfig:
    """
    如果安装了python-dotenv,支持从.env文件加载配置项

    !!! warning

        会从环境变量覆盖配置，列如access_token参数

        优先级为: 环境变量access_token > 环境变量ACCESS_TOKEN > ExecutorConfig().access_token

    """

    xxl_admin_baseurl: str
    """xxl-admin服务端暴露的restful接口url(如http://localhost:8080/xxl-job-admin/api/). 必填"""
    executor_app_name: str
    """xxl-admin上定义的执行器名称,必须一致否则无法注册(如xxl-job-executor-sample). 必填"""
    access_token: Optional[str] = None
    """调度器的token. Default: None"""

    executor_log_path: str = "pyxxl.log"
    """执行器日志输出的路径(注意路径必须存在). Default: pyxxl.log"""
    executor_port: int = 9999
    """执行器绑定的http服务的端口,作用同host. Default: 9999"""
    executor_host: str = field(default_factory=get_network_ip)
    """执行器绑定的host,xxl-admin通过这个host来回调pyxxl执行器,如果不填会默认取第一个网卡的地址. Default: 获取到第一个网卡的ip地址"""
    executor_listen_port: int = 0
    """Default: executor_port"""
    executor_listen_host: str = ""
    """
    执行器HTTP服务绑定的HOST,大部分情况下不需要设置. Default: executor_host

    当执行器通过了端口转发暴露给admin的时候,需要把executor_host填写为直连admin的地址.

    列如调用路径为 xxl-admin -> nginx_ip_or_domain:80 -> executor:9999
    这个时候需要配置为

        executor_port=80
        executor_host=nginx_ip_or_domain
        executor_listen_port=9999
        executor_listen_host="0.0.0.0"

    """

    max_workers: int = 30
    """执行器线程池（执行同步任务时使用）. Default: 30"""
    task_timeout: int = 60 * 10
    """任务的默认超时时间,如果调度器传了以参数executorTimeout为准. Default: 60 * 10"""
    task_queue_length: int = 30
    """任务的队列长度.单机串行的队列长度,当阻塞的任务大于此值时会抛弃. Default: 30"""
    graceful_close: bool = False
    """是否优雅关闭. Default: True"""
    graceful_timeout: int = 60 * 5
    """优雅关闭的等待时间,超过改时间强制停止任务. Default: 60 * 5"""

    log_target: Literal["disk", "redis"] = "disk"
    """任务日志存储的地方.  Default: disk"""
    log_local_dir: str = "logs"
    """任务日志存储的本地目录,默认为当前目录logs文件夹"""
    log_redis_uri: str = ""
    """任务日志存储到redis的连接地址"""
    log_expired_days: int = 14
    """任务日志存储的本地的过期天数. Default: 14"""

    dotenv_try: bool = True
    dotenv_path: Optional[str] = None
    """.env文件的路径,默认为当前路径下的.env文件."""
    debug: bool = False

    def __post_init__(self) -> None:
        if self.debug:
            setup_logging(self.executor_log_path, __name__, level=logging.DEBUG)

        if self.dotenv_try:
            self._try_load_from_dotenv()

        self._valid_xxl_admin_baseurl()
        self._valid_executor_app_name()
        self._valid_logger_target()

        if not self.executor_listen_host:
            self.executor_listen_host = self.executor_host

        if not self.executor_listen_port:
            self.executor_listen_port = self.executor_port

        logger.debug("init config: %s", asdict(self))

    def _try_load_from_dotenv(self) -> None:
        try:
            from dotenv import load_dotenv

            load_dotenv(self.dotenv_path)
        except ImportError:  # pragma: no cover
            pass

        for param in inspect.signature(ExecutorConfig).parameters.values():
            env_val = os.getenv(param.name) or os.getenv(param.name.upper())
            if env_val is not None:
                logger.debug("Get [%s] config from env: [%s]" % (param.name, env_val))
                if param.annotation is bool:
                    real_value = env_val in ["true", "True"]
                else:
                    real_value = param.annotation(env_val)
                setattr(self, param.name, real_value)

    def _valid_xxl_admin_baseurl(self) -> None:
        _admin_url: URL = URL(self.xxl_admin_baseurl)
        if not (_admin_url.scheme.startswith("http") and _admin_url.path.endswith("/")):
            raise ValueError("admin_url must like http://localhost:8080/xxl-job-admin/api/")

    def _valid_executor_app_name(self) -> None:
        if not self.executor_app_name:
            raise ValueError("executor_app_name is required.")

    def _valid_logger_target(self) -> None:
        if self.log_target == "disk" and not self.log_local_dir:
            raise ValueError("log_target 'disk' config item 'log_local_dir' is necessary.")

        if self.log_target == "redis" and not self.log_redis_uri:
            raise ValueError("log_target 'redis' config item 'log_redis_uri' is necessary.")

    @property
    def executor_baseurl(self) -> str:
        """暴露给xxl-admin的地址"""
        return "http://{host}:{port}".format(host=self.executor_host, port=self.executor_port)



import asyncio
from concurrent.futures import ThreadPoolExecutor
from dataclasses import asdict, is_dataclass
from typing import Any

from aiohttp import web
from prometheus_client import Counter, Gauge, Info
from prometheus_client.exposition import _bake_output
from prometheus_client.registry import REGISTRY

from pyxxl.ctx import g
from pyxxl.executor import Executor

FAILED_COUNTER = Counter("failed", "task failed number.", ["jobId", "reason"])
SUCCESS_COUNTER = Counter("success", "task success number.", ["jobId"])

RUNNING_TASKS = Gauge("running_tasks", "running tasks")
QUEUE_TASKS = Gauge("queue_tasks", "queue_tasks", ["jobId"])
ASYNCIO_TASKS_TOTAL = Gauge("asyncio_tasks_total", "ASYNCIO_TASKS_TOTAL")

RUNNING_TASK_INFO = Info("running_task", "running task info", ["pk"])
QUEUE_TASKS_INFO = Info("queue_task", "queue task info", ["pk"])
THREAD_POOL_INFO = Info("executor_thread_pool", "executor_thread_pool")

routes = web.RouteTableDef()


def success() -> None:
    SUCCESS_COUNTER.labels(g.xxl_run_data.jobId).inc(1)


def failed(reason: str) -> None:
    FAILED_COUNTER.labels(g.xxl_run_data.jobId, reason).inc(1)


def as_str_dict(obj: Any) -> dict:
    if is_dataclass(obj):
        obj = asdict(obj)
    return {k: str(v) for k, v in obj.items()}


def _get_thread_pool_info(pool: ThreadPoolExecutor) -> dict:
    data = {}
    data["wait_qsize"] = str(pool._work_queue.qsize())
    data["current_threads"] = str(len(pool._threads))
    data["max_workers"] = str(pool._max_workers)
    data["idle_threads"] = str(pool._idle_semaphore._value)  # type: ignore[attr-defined]
    return data


@routes.get("/metrics")
async def metrics(request: web.Request) -> web.Response:
    # init
    RUNNING_TASK_INFO.clear()
    QUEUE_TASKS_INFO.clear()
    ASYNCIO_TASKS_TOTAL.set(len(asyncio.all_tasks()))
    # export executor info
    executor: Executor = request.app["executor"]
    RUNNING_TASKS.set(len(executor.tasks))

    for k, v in executor.tasks.items():
        RUNNING_TASK_INFO.labels(k).info(as_str_dict(v.data))

    for kk, queue in executor.queue.items():
        QUEUE_TASKS.labels(kk).set(queue.qsize())
        QUEUE_TASKS_INFO.labels(kk).info({"detail": str(queue)})
    # thread pool
    THREAD_POOL_INFO.info(_get_thread_pool_info(executor.thread_pool))

    params = request.query
    # todo
    _, headers, output = _bake_output(REGISTRY, "", "", params, True)
    return web.Response(body=output, headers=headers)


def mount_app(app: web.Application) -> None:
    app.add_routes(routes)



import logging
from typing import TYPE_CHECKING

from aiohttp import web

from pyxxl import error
from pyxxl.schema import RunData
from pyxxl.utils import try_import

if TYPE_CHECKING:
    from pyxxl.logger import LogBase

logger = logging.getLogger(__name__)

routes = web.RouteTableDef()


@routes.post("/beat")
async def beat(request: web.Request) -> web.Response:
    logger.debug("beat")
    return web.json_response(dict(code=200, msg=None))


@routes.post("/idleBeat")
async def idle_beat(request: web.Request) -> web.Response:
    data = await request.json()
    job_id = data["jobId"]
    logger.debug("idleBeat: %s" % data)
    if await request.app["executor"].is_running(data["jobId"]):
        return web.json_response(dict(code=500, msg="job %s is running." % job_id))
    return web.json_response(dict(code=200, msg=None))


@routes.post("/run")
async def run(request: web.Request) -> web.Response:
    """
        {
        "jobId":1,                                  // 任务ID
        "executorHandler":"demoJobHandler",         // 任务标识
        "executorParams":"demoJobHandler",          // 任务参数
        "executorBlockStrategy":"COVER_EARLY",      // 任务阻塞策略，可选值参考 com.xxl.job.core.enums.ExecutorBlockStrategyEnum
        "executorTimeout":0,                        // 任务超时时间，单位秒，大于零时生效
        "logId":1,                                  // 本次调度日志ID
        "logDateTime":1586629003729,                // 本次调度日志时间
        "glueType":"BEAN",                          // 任务模式，可选值参考 com.xxl.job.core.glue.GlueTypeEnum
        "glueSource":"xxx",                         // GLUE脚本代码
        "glueUpdatetime":1586629003727,             // GLUE脚本更新时间，用于判定脚本是否变更以及是否需要刷新
        "broadcastIndex":0,                         // 分片参数：当前分片
        "broadcastTotal":0                          // 分片参数：总分片
    }
    """
    data = await request.json()
    run_data = RunData(**data)
    logger.info("Get task request. jobId=%s logId=%s [%s]" % (run_data.jobId, run_data.logId, run_data))
    msg = None
    try:
        msg = await request.app["executor"].run_job(run_data)
    except error.JobDuplicateError as e:
        return web.json_response(dict(code=500, msg=e.message))
    except error.JobNotFoundError as e:
        return web.json_response(dict(code=500, msg=e.message))

    return web.json_response(dict(code=200, msg=msg))


@routes.post("/kill")
async def kill(request: web.Request) -> web.Response:
    data = await request.json()
    await request.app["executor"].cancel_job(data["jobId"], include_queue=True)
    return web.json_response(dict(code=200, msg=None))


@routes.post("/log")
async def log(request: web.Request) -> web.Response:
    """
        {
        "logDateTim":0,     // 本次调度日志时间
        "logId":0,          // 本次调度日志ID
        "fromLineNum":0     // 日志开始行号，滚动加载日志
    }
    """
    data = await request.json()
    logger.debug("get log request %s" % data)
    executor_log: LogBase = request.app["executor_log"]
    response = {
        "code": 200,
        "msg": None,
        "content": await executor_log.get_logs(data),
    }
    return web.json_response(response)


def create_app() -> web.Application:
    app = web.Application()
    app.add_routes(routes)
    if try_import("prometheus_client"):
        from pyxxl.prometheus import mount_app

        mount_app(app)

    return app



import logging
import time
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, AsyncGenerator, Optional, Union

from pyxxl.types import LogRequest, LogResponse
from pyxxl.utils import STD_FORMATTER, try_import

from .common import MAX_LOG_TAIL_LINES, LogBase

if TYPE_CHECKING:
    from logging import Handler

    import redis
else:
    redis = try_import("redis")


KEY_PREFIX = "pyxxl:log:{app}:{log_id}"


class RedisHandler(logging.Handler):
    def __init__(
        self,
        key: str,
        ttl: int,
        rclient: "redis.Redis",
        *,
        level: int = logging.NOTSET,
        max_lines: Optional[int] = None,
    ) -> None:
        super().__init__(level)
        self.rclient = rclient
        self.key = key
        self.ttl = ttl
        self.max_lines = max_lines or MAX_LOG_TAIL_LINES

    def emit(self, record: Any) -> None:
        try:
            p = self.rclient.pipeline()
            p.rpush(self.key, self.format(record))
            p.ltrim(self.key, -self.max_lines, -1)
            p.expire(self.key, self.ttl)
            p.execute()
        except redis.RedisError as e:  # pragma: no cover
            print("log to redis failed. %s" % str(e))  # pragma: no cover


class RedisLog(LogBase):
    def __init__(
        self,
        app: str,
        redis_client: Union[str, "redis.ConnectionPool"],
        log_tail_lines: int = 0,
        expired_days: int = 14,
    ) -> None:
        if redis is None:
            raise ImportError("Depend on redis. pip install redis or pip install pyxxl[redis].")  # pragma: no cover

        self.app = app
        self.log_tail_lines = log_tail_lines or MAX_LOG_TAIL_LINES
        self.expired_days = expired_days
        if isinstance(redis_client, str):
            rclient = redis.Redis.from_url(redis_client)
        elif isinstance(redis_client, redis.ConnectionPool):
            rclient = redis.Redis(connection_pool=redis_client)
        else:
            raise TypeError(
                "pool expect Union[str, redis.ConnectionPool], got %s." % type(redis_client)
            )  # pragma: no cover
        self.rclient = rclient

    def get_logger(self, log_id: int, *, stdout: bool = True, level: int = logging.INFO) -> logging.Logger:
        logger = logging.getLogger("pyxxl-task-{%s}" % log_id)
        logger.propagate = False
        logger.setLevel(level)
        handlers: list[Handler] = [logging.StreamHandler()] if stdout else []
        handlers.append(RedisHandler(self.key(log_id), self.expired_days * 3600 * 24, self.rclient))
        for h in handlers:
            h.setFormatter(STD_FORMATTER)
            h.setLevel(level)
            logger.addHandler(h)
        return logger

    def key(self, log_id: int) -> str:
        return KEY_PREFIX.format(app=self.app, log_id=log_id)

    async def read_task_logs(self, log_id: int, *, key: str = None) -> str:
        key = key or self.key(log_id)
        # todo: use async
        return "".join(i.decode() for i in self.rclient.lrange(key, 0, -1))

    async def get_logs(self, request: LogRequest, *, key: str = None) -> LogResponse:
        key = key or self.key(request["logId"])
        from_line = request["fromLineNum"] - 1
        to_line = request["fromLineNum"] - 1 + self.log_tail_lines
        llen = self.rclient.llen(key)
        if from_line >= llen:
            logs = "No such logid logs." if llen == 0 else ""
            to_line_num = request["fromLineNum"]
        else:
            # lrange 0 20   [0, 20]
            logs = "".join(i.decode() for i in self.rclient.lrange(key, from_line, to_line - 1))
            to_line_num = to_line

        return LogResponse(
            fromLineNum=request["fromLineNum"],
            toLineNum=to_line_num,
            logContent=logs,
            isEnd=llen <= to_line,
        )

    @asynccontextmanager
    async def mock_write(self, *lines: Any) -> AsyncGenerator[str, None]:
        key = self.key(int(time.time() * 1000))
        self.rclient.rpush(key, *lines)
        yield key
        self.rclient.delete(key)

    @asynccontextmanager
    async def mock_logger(self, log_id: int) -> AsyncGenerator[LogBase, None]:
        yield self
        self.rclient.delete(self.key(log_id))



from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class RunData:
    """
    调度器传入的所有参数，执行函数通过g来获取这些参数

    !!! example

        ```python
        from pyxxl.ctx import g

        @xxxxx
        async def test():
            print(g.xxl_run_data.logId)
        ```

    """

    jobId: int
    logId: int
    executorHandler: str
    executorBlockStrategy: str

    executorParams: Optional[str] = None
    executorTimeout: Optional[int] = None
    logDateTime: Optional[int] = None
    glueType: Optional[str] = None
    glueSource: Optional[str] = None
    glueUpdatetime: Optional[int] = None
    broadcastIndex: Optional[int] = None
    broadcastTotal: Optional[int] = None



import pytest

from pyxxl.ctx import g
from pyxxl.enum import executorBlockStrategy
from pyxxl.executor import Executor
from pyxxl.schema import RunData


@pytest.mark.asyncio
async def test_runner_callback(executor: Executor):
    @executor.handler.register
    async def test_ctx():
        logId = g.xxl_run_data.logId
        assert logId == 1

    @executor.handler.register
    def test_ctx_sync():
        logId = g.xxl_run_data.logId
        assert logId == 1

    for handler in ["test_ctx", "test_ctx_sync"]:
        data = RunData(
            **dict(
                logId=1,
                jobId=11,
                executorHandler=handler,
                executorBlockStrategy=executorBlockStrategy.SERIAL_EXECUTION.value,
            )
        )
        await executor.run_job(data)
        await executor.graceful_close()
        assert executor.xxl_client.callback_result.get(1) == 200



import asyncio

import pytest
from aiohttp.test_utils import TestClient

from pyxxl.utils import try_import

from .test_server import send_demoJobHandler


@pytest.mark.asyncio
@pytest.mark.skipif(not try_import("prometheus_client"), reason="不存在prometheus_client")
async def test_metrics(cli: TestClient):
    for _ in range(3):
        await send_demoJobHandler(cli, jobId=630)
        await send_demoJobHandler(cli, jobId=631, executorHandler="demoJobHandlerSync")
        await asyncio.sleep(0.01)
    await asyncio.sleep(1)
    resp = await cli.get("/metrics")
    assert resp.status == 200
    assert "python_gc_objects_collected_total" in await resp.text()



import asyncio
import time
from typing import Any, Generator

import pytest
import pytest_asyncio
from aiohttp.web import Application
from pytest_aiohttp.plugin import AiohttpClient, TestClient

from pyxxl import ExecutorConfig
from pyxxl.executor import Executor
from pyxxl.tests.utils import INSTALL_REDIS, REDIS_TEST_URI, MokePyxxlRunner, MokeXXL

GLOBAL_JOB_ID = 1
GLOBAL_CONFIG: Any = dict(
    xxl_admin_baseurl="http://localhost:8080/xxl-job-admin/api/",
    executor_app_name="xxl-job-executor-sample",
    executor_host="127.0.0.1",
    graceful_close=False,
    task_queue_length=5,
    dotenv_try=False,
)

xxl_admin_baseurl = "http://localhost:8080/xxl-job-admin/api/"


def _create_job_id() -> int:
    global GLOBAL_JOB_ID
    GLOBAL_JOB_ID += 1
    return GLOBAL_JOB_ID


@pytest.fixture(scope="session")
def event_loop() -> Generator:
    """Create an instance of the default event loop for each test case."""
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(
    scope="session",
    params=[
        ExecutorConfig(**GLOBAL_CONFIG),
        pytest.param(
            ExecutorConfig(**GLOBAL_CONFIG, log_target="redis", log_redis_uri=REDIS_TEST_URI),
            marks=pytest.mark.skipif(not INSTALL_REDIS, reason="no redis package."),
        ),
    ],
    ids=["disk", "redis"],
)
def executor(request: Any) -> Executor:
    return Executor(MokeXXL("http://localhost:8080/xxl-job-admin/api/"), request.param, handler=None)


@pytest.fixture(scope="session")
def web_app(executor: Executor) -> Application:
    runner = MokePyxxlRunner(executor.config)

    @runner.register(name="demoJobHandler")
    async def test_task() -> None:
        await asyncio.sleep(10)

    @runner.register(name="demoJobHandlerSync")
    def test_task_sync() -> None:
        time.sleep(5)

    return runner.create_server_app()


@pytest_asyncio.fixture
async def cli(aiohttp_client: AiohttpClient, web_app: Application) -> TestClient:
    return await aiohttp_client(web_app)


@pytest.fixture(scope="function")
def job_id() -> int:
    return _create_job_id()



import pytest
from aiohttp import web
from pytest_aiohttp.plugin import AiohttpClient

from pyxxl.xxl_client import XXL


@pytest.mark.asyncio
async def test_param_admin_url():
    XXL("http://localhost:8080/xxl-job-admin/api/")
    XXL("https://localhost:8080/xxl-job-admin/api/")

    with pytest.raises(ValueError):
        XXL("htp://localhost:8080/xxl-job-admin/api/")

    with pytest.raises(ValueError):
        XXL("http://localhost:8080/xxl-job-admin/api")


@pytest.mark.asyncio
async def test_client(aiohttp_client: AiohttpClient) -> None:
    async def moke_registry_api(request: web.Request):
        data = await request.json()
        if data.get("registryKey") == "server_test":
            return web.json_response({"code": 500, "msg": "1"}, status=500)

        if data.get("registryKey") == "status_test":
            return web.json_response({"code": 500, "msg": "1"}, status=200)

        return web.json_response({"code": 200, "msg": "1"})

    async def moke_callback_api(request: web.Request):
        return web.json_response({"code": 200, "msg": "1"})

    app = web.Application()
    app.router.add_post("/xxl-job-admin/api/registry", moke_registry_api)
    app.router.add_post("/xxl-job-admin/api/callback", moke_callback_api)
    session = await aiohttp_client(app)
    xxl_client = XXL("http://localhost:8080/xxl-job-admin/api/", session=session)
    # registry
    assert await xxl_client.registry("key", "value")
    assert not (await xxl_client.registry("server_test", "value"))
    assert not (await xxl_client.registry("status_test", "value"))
    # callback
    await xxl_client.callback(123, 123123123)



import asyncio
import logging
from abc import ABC, abstractmethod
from typing import Any, AsyncContextManager, Optional

from pyxxl.types import LogRequest, LogResponse

logger = logging.getLogger(__name__)


MAX_LOG_TAIL_LINES = 1000


class LogBase(ABC):
    @abstractmethod
    def get_logger(self, log_id: int, *, stdout: bool = True, level: int = logging.INFO) -> logging.Logger:
        ...

    @abstractmethod
    async def get_logs(self, request: LogRequest, *, key: Optional[str] = None) -> LogResponse:
        ...

    @abstractmethod
    async def read_task_logs(self, log_id: int, *, key: Optional[str] = None) -> str:
        """一次性读取某个log id的所有日志,主要用于单测"""
        ...

    @abstractmethod
    def mock_write(self, *lines: Any) -> AsyncContextManager[str]:
        ...

    @abstractmethod
    def mock_logger(self, log_id: int) -> AsyncContextManager["LogBase"]:
        ...

    async def expired_once(self) -> None:  # noqa: B027
        """执行一次批量过期操作,如果是redis啥的自带过期就无需实现此方法"""
        pass

    async def expired_loop(self, seconds: int = 3600) -> None:
        """
        Args:
            seconds (int, optional): one loop seconds. Defaults to 3600.
        """
        logger.debug("start expired_loop...")
        while True:
            await self.expired_once()
            await asyncio.sleep(seconds)



import asyncio
import functools
import logging
from multiprocessing import Process
from typing import Any, AsyncGenerator, Optional

from aiohttp import web

from pyxxl.executor import Executor, JobHandler
from pyxxl.logger import DiskLog, LogBase, RedisLog
from pyxxl.server import create_app
from pyxxl.setting import ExecutorConfig
from pyxxl.utils import setup_logging, try_import
from pyxxl.xxl_client import XXL

logger = logging.getLogger(__name__)

if try_import("prometheus_client"):
    from pyxxl import prometheus

    Executor: Executor = functools.partial(  # type: ignore[no-redef]
        Executor,
        successed_callback=prometheus.success,
        failed_callback=prometheus.failed,
    )


class PyxxlRunner:
    daemon: Optional[Process] = None

    def __init__(
        self,
        config: ExecutorConfig,
        handler: Optional[JobHandler] = None,
    ):
        """
        !!! example

            ```python
            runner = PyxxlRunner(
                ExecutorConfig(
                    xxl_admin_baseurl="http://localhost:8080/xxl-job-admin/api/",
                    executor_app_name="xxl-job-executor-sample",
                    executor_host="172.17.0.1"
                    )
                ,
                handler=xxl_handler,
            )
            ```
        Args:
            config (ExecutorConfig): 配置参数
            handler (JobHandler, optional): 执行器支持的job,没有预先定义的job名称也会执行失败
        """

        self.handler = handler or JobHandler()
        self.config = config
        log_level = logging.DEBUG if self.config.debug else logging.INFO
        setup_logging(self.config.executor_log_path, "pyxxl", level=log_level)

    async def _register_task(self, xxl_client: XXL) -> None:
        # todo: 这是个调度器的bug，必须循环去注册，不然会显示为离线
        # https://github.com/xuxueli/xxl-job/issues/2090
        try:
            while True:
                await xxl_client.registry(self.config.executor_app_name, self.config.executor_baseurl)
                await asyncio.sleep(10)
        finally:
            logger.warning("Register task is exit.")

    def _get_xxl_clint(self) -> XXL:
        """for moke"""
        return XXL(self.config.xxl_admin_baseurl, token=self.config.access_token)

    def _get_log(self) -> LogBase:
        if self.config.log_target == "disk":
            return DiskLog(log_path=self.config.log_local_dir, expired_days=self.config.log_expired_days)

        if self.config.log_target == "redis":
            return RedisLog(
                self.config.executor_app_name, self.config.log_redis_uri, expired_days=self.config.log_expired_days
            )

        raise NotImplementedError

    async def _cleanup_ctx(self, app: web.Application) -> AsyncGenerator:
        # setup task log
        executor_log = self._get_log()
        executor_log_task = asyncio.create_task(executor_log.expired_loop(), name="log_task")

        xxl_client = self._get_xxl_clint()
        executor = Executor(xxl_client, config=self.config, handler=self.handler, logger_factory=executor_log)
        register_task = asyncio.create_task(self._register_task(xxl_client), name="register_task")

        app["xxl_client"] = xxl_client
        app["executor"] = executor
        app["executor_log"] = executor_log

        if executor.handler:
            logger.info("register with handlers %s", list(executor.handler.handlers_info()))
        else:
            logger.warning("register with handlers is empty")  # pragma: no cover

        yield

        register_task.cancel()
        executor_log_task.cancel()
        await xxl_client.registryRemove(self.config.executor_app_name, self.config.executor_baseurl)
        if self.config.graceful_close:
            await executor.graceful_close(self.config.graceful_timeout)
        else:
            await executor.shutdown()
        await xxl_client.close()
        logger.info("cleanup executor success.")

    def create_server_app(self) -> web.Application:
        """获取执行器的app对象,可以使用自己喜欢的服务器启动这个webapp"""
        app = create_app()
        app.cleanup_ctx.append(self._cleanup_ctx)
        return app

    def run_executor(self, handle_signals: bool = True) -> None:
        """用aiohttp的web服务器启动执行器"""
        web.run_app(
            self.create_server_app(),
            port=self.config.executor_listen_port,
            host=self.config.executor_listen_host,
            handle_signals=handle_signals,
        )

    def _runner(self) -> None:
        self.run_executor(handle_signals=True)

    def run_with_daemon(self) -> None:
        """新开一个进程以后台方式运行,一般和gunicorn一起使用"""

        daemon = Process(target=self._runner, name="pyxxljob", daemon=True)
        daemon.start()
        self.daemon = daemon

    @property
    def register(self) -> Any:
        return self.handler.register

    # def exit_daemon(self):
    #     logger.info("Exit daemon name=%s", self.daemon.name )
    #     self.daemon.terminate()



import importlib
import logging
import platform
import socket
from logging.handlers import RotatingFileHandler
from typing import Any, List, Optional

from pyxxl.ctx import g

DEFAULT_FORMAT = (
    "%(asctime)s.%(msecs)03d [%(threadName)s] [%(logId)s] "
    "%(levelname)s %(pathname)s(%(funcName)s:%(lineno)d) - %(message)s"
)
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
STD_FORMATTER = logging.Formatter(DEFAULT_FORMAT, datefmt=DATE_FORMAT)
DEFAULT_FILE_SIZE = 50 * 1024 * 1024
DEFAULT_BACKUP_FILE_COUNT = 5


def get_network_ip() -> str:
    """获取本机地址,会获取首个网络地址"""
    if platform.system() == "Darwin":
        return "127.0.0.1"  # todo
    else:
        _, _, ipaddrlist = socket.gethostbyname_ex(socket.gethostname())
    return ipaddrlist[0]


def _init_log_record_factory() -> None:
    old_factory = logging.getLogRecordFactory()

    def _record_factory(*args: Any, **kwargs: Any) -> logging.LogRecord:
        record: Any = old_factory(*args, **kwargs)
        xxl_kwargs = g.try_get_data()
        record.logId = xxl_kwargs.logId if xxl_kwargs else "NotInTask"
        return record

    logging.setLogRecordFactory(_record_factory)


def setup_logging(
    path: str,
    name: str,
    level: int = logging.INFO,
    custom_handlers: Optional[List[logging.Handler]] = None,
    std_formatter: Optional[logging.Formatter] = None,
) -> logging.Logger:
    logger = logging.getLogger(name)
    if logger.handlers:
        return logger

    std_formatter = std_formatter or STD_FORMATTER

    _init_log_record_factory()

    logger.setLevel(level)
    logger.propagate = False

    handlers: List[logging.Handler] = [
        logging.StreamHandler(),
        RotatingFileHandler(path, maxBytes=DEFAULT_FILE_SIZE, backupCount=DEFAULT_BACKUP_FILE_COUNT, delay=True),
    ]
    if custom_handlers:
        handlers.extend(custom_handlers)

    for h in handlers:
        h.setFormatter(std_formatter)
        h.setLevel(level)
        logger.addHandler(h)
    return logger


def try_import(module: str) -> Optional[Any]:
    try:
        return importlib.import_module(module)
    except ImportError:
        pass
    return None



import logging
import os
import time
import weakref
from contextlib import asynccontextmanager
from logging import FileHandler
from pathlib import Path
from typing import TYPE_CHECKING, Any, AsyncGenerator, List

import aiofiles

from pyxxl.types import LogRequest, LogResponse
from pyxxl.utils import STD_FORMATTER

from .common import LogBase

if TYPE_CHECKING:
    from logging import Handler


LOG_NAME_PREFIX = "pyxxl-{log_id}.log"
LOG_NAME_REGEX = "pyxxl-*.log"
MAX_LOG_TAIL_LINES = 1000
logger = logging.getLogger(__name__)


def _close_file_stream(_logger: logging.Logger) -> None:
    # !!! StreamHandler remove会有问题，需要判断是否是FileHandler
    for h in _logger.handlers:
        if isinstance(h, logging.FileHandler):
            logger.debug("close file log object: {}.".format(h))
            h.close()
            _logger.removeHandler(h)


class XXLogger:
    """为了可以关闭文件写入流"""

    def __init__(self, logger: logging.Logger) -> None:
        self._logger = logger

        self._finalizer = weakref.finalize(self, _close_file_stream, self._logger)

    def remove(self) -> None:
        self._finalizer()

    @property
    def removed(self) -> bool:
        return not self._finalizer.alive

    def __getattr__(self, name: str) -> Any:
        if name in ["info", "error", "warning", "debug", "exception", "handlers"]:
            return getattr(self._logger, name)


class DiskLog(LogBase):
    def __init__(self, log_path: str, log_tail_lines: int = 0, expired_days: int = 14) -> None:
        self.log_path = Path(log_path)
        if not self.log_path.exists():
            self.log_path.mkdir()  # pragma: no cover
            logger.info("create logdir %s" % self.log_path)  # pragma: no cover
        self.log_tail_lines = log_tail_lines or MAX_LOG_TAIL_LINES
        self.expired_days = expired_days

    def key(self, log_id: int) -> str:
        return self.log_path.joinpath(LOG_NAME_PREFIX.format(log_id=log_id)).absolute().as_posix()

    def get_logger(  # type:ignore[override]
        self, log_id: int, *, stdout: bool = True, level: int = logging.INFO
    ) -> XXLogger:
        logger = logging.getLogger("pyxxl-task-{%s}" % log_id)
        logger.propagate = False
        logger.setLevel(level)
        handlers: list[Handler] = [logging.StreamHandler()] if stdout else []
        handlers.append(FileHandler(self.key(log_id), delay=True))
        for h in handlers:
            h.setFormatter(STD_FORMATTER)
            h.setLevel(level)
            logger.addHandler(h)
        return XXLogger(logger)

    async def get_logs(self, request: LogRequest, *, key: str = None) -> LogResponse:
        # todo: 优化获取中间行的逻辑，缓存之前每行日志的大小然后直接seek
        logs = ""
        to_line_num = request["fromLineNum"]  # start with 1
        is_end = False
        key = key or self.key(request["logId"])
        try:
            async with aiofiles.open(key, mode="r") as f:
                for i in range(1, request["fromLineNum"] + self.log_tail_lines):
                    log = await f.readline()
                    if log == "":
                        is_end = True
                        break
                    elif i >= request["fromLineNum"]:
                        to_line_num = i
                        logs += log
        except FileNotFoundError as e:
            logger.warning(str(e), exc_info=True)
            logs = "No such logid logs."

        return LogResponse(
            fromLineNum=request["fromLineNum"],
            toLineNum=to_line_num,
            logContent=logs,
            isEnd=is_end,
        )

    async def read_task_logs(self, log_id: int, *, key: str = None) -> str:
        key = key or self.key(log_id)
        async with aiofiles.open(key, mode="r") as f:
            return await f.read()

    async def expired_once(self) -> None:
        now = time.time()
        expire_timestamp = now - 3600 * 24 * self.expired_days
        del_list: List[Path] = []
        if self.log_path.exists():
            for sub_path in [i for i in self.log_path.glob(LOG_NAME_REGEX) if i.is_file()]:
                ctime = os.path.getctime(sub_path.absolute())
                if ctime < expire_timestamp:
                    del_list.append(sub_path)

        if del_list:
            logger.info("delete expired logs [{}] - {}".format(len(del_list), " | ".join(str(i) for i in del_list)))
            for i in del_list:
                i.unlink()

    @asynccontextmanager
    async def mock_write(self, *lines: Any) -> AsyncGenerator[str, None]:
        async with aiofiles.tempfile.NamedTemporaryFile() as f:
            await f.writelines(lines)
            await f.flush()
            await f.seek(0)
            yield str(f.name)

    @asynccontextmanager
    async def mock_logger(self, _log_id: int) -> AsyncGenerator[LogBase, None]:
        async with aiofiles.tempfile.TemporaryDirectory() as d:
            handler = DiskLog(log_path=d)
            yield handler
